# import_ifp.py @space_view3d_gta_tools
# 2011 ponz
script_info = "GTA IFP Exporter ( built:2011.12.11 )"

# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or ( at your option ) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

#####
# ToDo.
#  - NLA support
#####

import bpy
import os
import time
import struct
from bpy.props import *
from mathutils import *
from math import *

## Constants


## Classes
# Data Class
class ClassIFPFram:
	def __init__( self ):
		self.time      = -1
		self.rot       = []
		self.pos       = []
		self.scale     = []

class ClassIFPObj:
	def __init__( self ):
		self.bid      = -1
		self.name     = ""
		self.kfrm     = ""
		self.frams    = []

class ClassIFPAnim:
	def __init__( self ):
		self.name     = ""
		self.objs     = []


## Functions
def GetBone( pose_bones, bone_id ):  # "bone_id" is a custom property which set in importing DFF
	for pose_bone in pose_bones:
		if bone_id == pose_bone.bone["bone_id"]:
			return pose_bone
	return None

def GetBoneByName( pose_bones, name ):
	for pose_bone in pose_bones:
		if name == pose_bone.name:
			return pose_bone
	return None

def GetFCurve( action, target, data_path, array_index, new_flg ):
	## new_flg: if fc is not exist,
	##   False : return None
	##   True  : return New FCurve
	attr = target.path_from_id( data_path )
	for fc in action.fcurves:
		if fc.data_path == attr:
			if fc.array_index == array_index:
				return fc
	if new_flg:
		if target.id_data == target:
			if   data_path == "location":
				target_name = "Location"
			elif data_path == "rotation_quaternion":
				target_name = "Rotation"
		else:
			target_name = target.name
		return action.fcurves.new( attr, index = array_index, action_group = target_name )
	else:
		return None

def GetSingleKey( target, data_path, array_index, target_frame ):
	current_frame = bpy.context.scene.frame_current
	bpy.context.scene.frame_set( target_frame )
	if "location" == data_path:
		single_key = target.location[array_index]
	else:
		single_key = target.rotation_quaternion[array_index]
	bpy.context.scene.frame_set( current_frame )
	return single_key

def ExportIFP():
	time_ini=time.clock()
	
	ifp_props=bpy.context.scene.gta_tools.ifp_props
	global script_info
	print( "\n-----\n" + script_info + "\n-----" )
	
	## Get Selected Armature Data
	aobj=bpy.context.active_object
	if aobj.mode != 'POSE':
		print( "-----\nMode Error: "+aobj.name+" (mode: "+aobj.mode+")" )
		return
	print( "-----\nTarget Object: "+aobj.name+" (mode: "+aobj.mode+")" )
	arm=aobj.data
	
	if aobj.type != 'ARMATURE':
		print( "-----\nObject Type Error: "+aobj.data.name+" (type: "+aobj.type+")" )
		return
	print( "Object Type: "+aobj.data.name+" (type: "+aobj.type+")" )
	
	pose=aobj.pose
	pose_bones=pose.bones
	print( "Number of Pose Bones: "+str( len( pose_bones ) ) )
	
	if aobj.animation_data:
		action = aobj.animation_data.action
	else:
		print( "-----\nError : No Anim Data " )
		return
	
	## Frame Range Setup
	scene_start = bpy.context.scene.frame_start
	scene_end   = bpy.context.scene.frame_end
	print( "Anim Start Frame : ", action.frame_range[0] )
	print( "Anim End Frame   : ", action.frame_range[1] )
	print( "-----" )
	time_ofs = action.frame_range[0]
	if ifp_props.scene_range_trim:
		print( "Trim Export Anim Range as Scene Renge" )
		print( "  Scene Start Frame : ", scene_start )
		print( "  Scene End Frame   : ", scene_end   )
		print( "-----" )
		time_ofs = scene_start
	
	
	## Get Fcurves
	ifp_anim = ClassIFPAnim()
	ifp_anim.name = ifp_props.exp_anim_name
	
	for pose_bone in pose_bones:
		target = pose_bone
		bone_quat = pose_bone.bone.matrix.to_quaternion()
		
		fram_list = []
		key_flg = False
		pos_flg = False
		pos_dict_array = [ {}, {}, {} ]
		rot_dict_array = [ {}, {}, {}, {} ]
		
		## POS Key
		data_path = "location"
		for ixyz in range( 3 ):
			array_index = ixyz
			fc = GetFCurve( action, target, data_path, array_index, False )
			if fc:
				pos_flg = True
				key_flg = True
				for kf in fc.keyframe_points:
					if ifp_props.scene_range_trim:
						if scene_start > kf.co[0] or scene_end < kf.co[0]:
							continue
					fram_time = kf.co[0]
					fram_list.append( fram_time )
					pos_dict_array[ixyz][fram_time] = kf.co[1]
				
				if ifp_props.scene_range_trim:
					if 0 < len( fram_list ):
						fram_list.append( scene_start )
						fram_list.append( scene_end )
		
		
		## ROT Key
		data_path = "rotation_quaternion"
		for iwxyz in range( 4 ):
			array_index = iwxyz
			fc = GetFCurve( action, target, data_path, array_index, False )
			if fc:
				key_flg = True
				for kf in fc.keyframe_points:
					if ifp_props.scene_range_trim:
						if scene_start > kf.co[0] or scene_end < kf.co[0]:
							continue
					fram_time = kf.co[0]
					fram_list.append( fram_time )
					rot_dict_array[iwxyz][fram_time] = kf.co[1]
				
				if ifp_props.scene_range_trim:
					if 0 < len( fram_list ):
						print( scene_start, scene_end )
						fram_list.append( scene_start )
						fram_list.append( scene_end )
		
		if key_flg:
			fram_list  = sorted( set( fram_list ), key=fram_list.index ) ## make fram_list unique ( using "set" container )
			
			## Set IFP Data Structure
			ifp_obj = ClassIFPObj()
			if ifp_props.rev_bone:
				ifp_obj.name =  pose_bone.bone["org_name"]
			else:
				ifp_obj.name = pose_bone.name
			ifp_obj.bid  = pose_bone.bone["bone_id"]
			ifp_obj.kfrm = "KR00"
			
			for fram in fram_list:
				ifp_fram = ClassIFPFram()
				ifp_fram.time = fram - time_ofs
				tmp_quat = Quaternion( ( 1, 0, 0, 0 ) )
				tmp_vec  = Vector( ( 0, 0, 0 ) )
				
				if pos_flg:
					ifp_obj.kfrm = "KRT0"
					for ixyz in range( 3 ):
						if fram in pos_dict_array[ixyz]:
							tmp_vec[ixyz] = pos_dict_array[ixyz][fram]
						else:
							tmp_vec[ixyz] = GetSingleKey( target, "location", ixyz, fram )
					tmp_vec.rotate( bone_quat )
					ifp_fram.pos = list( tmp_vec )
				
				for iwxyz in range( 4 ):
					if fram in rot_dict_array[iwxyz]:
						tmp_quat[iwxyz] = rot_dict_array[iwxyz][fram]
					else:
						tmp_quat[iwxyz] = GetSingleKey( target, "rotation_quaternion", iwxyz, fram )
				
				ifp_fram.rot = list( bone_quat * tmp_quat )
				ifp_obj.frams.append( ifp_fram )
			
			print( " ", ifp_obj.bid, ", ", ifp_obj.kfrm, ", ", ifp_obj.name )
			### for Debug
			#print( ifp_obj.name, ifp_obj.bid, ifp_obj.kfrm )
			#for ifp_fram in ifp_obj.frams:
			#	print( " ", ifp_fram.time )
			#	if "KRT0" == ifp_obj.kfrm:
			#		print( "  ", ifp_fram.pos[0], ifp_fram.pos[1],  ifp_fram.pos[2] )
			#	print( "  ", ifp_fram.rot[0], ifp_fram.rot[1], ifp_fram.rot[2],  ifp_fram.rot[3] )
			
			ifp_anim.objs.append( ifp_obj )
	
	
	##########
	## Set IFP Data Sequence
	
	## AnimData
	anim = []
	framdata_size = 0
	for ifp_obj in ifp_anim.objs:
		## ObjData
		obj = []
		for ifp_fram in ifp_obj.frams:
			## FrameData
			# 2b   - Int16    - Quaternion X
			# 2b   - Int16    - Quaternion Y
			# 2b   - Int16    - Quaternion Z
			# 2b   - Int16    - Quaternion W
			# 2b   - Int16    - Time (in seconds)
			# 2b   - Int16    - Translation X
			# 2b   - Int16    - Translation Y
			# 2b   - Int16    - Translation Z
			fram = []
			framdata_size += 10
			fram.extend( struct.pack( "<5h", int( ifp_fram.rot[1]*4096 ), int( ifp_fram.rot[2]*4096 ), int( ifp_fram.rot[3]*4096 ), int( ifp_fram.rot[0]*4096 ), int( ifp_fram.time*2 ) ) )
			if "KRT0" == ifp_obj.kfrm:
				framdata_size += 6
				fram.extend( struct.pack( "<3h", int( ifp_fram.pos[0]*1024 ), int( ifp_fram.pos[1]*1024 ), int( ifp_fram.pos[2]*1024 ) ) )
			obj.extend( fram )
		
		## ObjData Header
		# 24b  - Char[24] - Object Name
		# 4b   - Int32    - Frame type: Child=3, Root=4
		# 4b   - Int32    - Number of Frames
		# 4b   - Int32    - Bone ID
		obj_name = ifp_obj.name
		if "KRT0" == ifp_obj.kfrm:
			fram_type = 4
		else:
			fram_type = 3
		num_frams = len( ifp_obj.frams )
		bone_id = ifp_obj.bid
		anim.extend( struct.pack( "<24s3I", bytes( obj_name.encode() ), fram_type, num_frams, bone_id ) )
		
		## Add ObjData
		anim.extend( obj )
	
	## AnimData Header
	# 24b  - Char[24] - Animation Name
	# 4b   - Int32    - Number of Objects
	# 4b   - Int32    - Size of frame data
	# 4b   - Int32    - Unknown, always 1
	animdata = []
	anim_name    = ifp_anim.name
	num_objs     = len( ifp_anim.objs )
	sizeof_frams = framdata_size
	unk = 1
	animdata.extend( struct.pack( "<24s3I", bytes( anim_name.encode() ), num_objs, sizeof_frams, unk ) )
	
	## Add AnimData
	animdata.extend( anim )
	
	## ANP3 Header
	# 4b   - FourCC   - 'ANP3' (Animation Package 3, Version identifier. However there is no pack with ANP2)
	# 4b   - Int32    - Offset to end of file
	# 24b  - Char[24] - internal file name used in the script
	# 4b   - Int32    - Number of Animations		
	ifp = []
	fourcc = "ANP3"
	ofs = 28 + len( animdata )
	internal_fn = ifp_props.exp_internal_file_name
	num_anims = 1
	ifp.extend( struct.pack( "<4sI24sI", bytes( fourcc.encode() ), ofs, bytes( internal_fn.encode() ), num_anims ) )
	
	## Add AnimData
	ifp.extend( animdata )
	
	
	##########
	## Write to ifp file
	
	## Open File
	#filepath = "F:\\data\\3DCG\\blender\\ifp_test\\test.ifp"
	filepath = ifp_props.exp_filepath
	try:
		file = open( filepath, "wb" )
	except:
		print( "-----\nError : failed to open " + filepath )
		return
	
	## Write Data
	file.write( struct.pack('<%dB' %( len( ifp ) ), *ifp ) )
	file.close()
	
	
	print( "-----\nFinished : Successfully." )
	print( "Elapsed Time : "+str( round( time.clock() - time_ini, 3 ) )+" sec\n-----\n" )
	return



