# import_ifp.py @space_view3d_gta_tools
# 2011 ponz
script_info = "GTA IFP Importer ( built:2011.12.11 )"

# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or ( at your option ) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

#####
# ToDo.
#  - NLA support
#####

import bpy
import os
import time
import struct
from bpy.props import *
from mathutils import *
from math import *

## Constants


## Classes
# Data Class
class ClassIFPFram:
	def __init__( self ):
		self.time      = -1
		self.rot       = []
		self.pos       = []
		self.scale     = []

class ClassIFPObj:
	def __init__( self ):
		self.bid      = -1
		self.name     = ""
		self.kfrm     = ""
		self.frams    = []

class ClassIFPAnim:
	def __init__( self ):
		self.name     = ""
		self.objs     = []


## Functions
def GetBone( pose_bones, bone_id ):  # "bone_id" is a custom property which set in importing DFF
	for pose_bone in pose_bones:
		if bone_id == pose_bone.bone["bone_id"]:
			return pose_bone
	return None

def GetBoneByName( pose_bones, name ):
	for pose_bone in pose_bones:
		if name == pose_bone.name:
			return pose_bone
	return None

def ResetPose():
	aobj=bpy.context.active_object
	pose=aobj.pose
	pose_bones=pose.bones
	#print( "Number of Pose Bones: "+str( len( pose_bones ) ) )
	quat_ini=Quaternion( [1,0,0,0] )
	co_ini=Vector( [0,0,0] )
	for pose_bone in pose_bones:
		pose_bone.rotation_quaternion=quat_ini
		pose_bone.location=co_ini
	aobj.rotation_mode='QUATERNION'
	

def ResetAnim():
	## Reset Anim Data
	aobj=bpy.context.active_object
	pose=aobj.pose
	pose_bones=pose.bones
	
	## Delete Keys
	bpy.context.scene.frame_set( 1 )
	if None != aobj.animation_data:
		action = aobj.animation_data.action
		fcurves = action.fcurves
		for fcurve in fcurves[:]:
			fcurves.remove(fcurve)
		action.user_clear()
		bpy.data.actions.remove(action)
	aobj.animation_data_clear()
	
	## Delete Un-Used Anim Data
	#for action in bpy.data.actions[:]:
	#	print(action)
	#	try:
	#		bpy.data.actions.remove(action)
	#		print(  "remove" )
	#	except:
	#		print(  "used" )

def ResetArmature():
	aobj=bpy.context.active_object
	aobj.rotation_mode='QUATERNION'
	aobj.rotation_quaternion=Quaternion( [1,0,0,0] )
	aobj.location=Vector( [0,0,0] )
	aobj.matrix_local=aobj.matrix_basis

def AnimDirection():
	aobj=bpy.context.active_object
	pose=aobj.pose
	pose_bones=pose.bones
	
	root_pose_bone=GetBone( pose_bones, 0 )
	root_quat = Quaternion( ( sqrt(1/2), 0, 0, sqrt(1/2) ) )
	root_quat = root_pose_bone.bone.matrix.to_quaternion().inverted() * root_quat
	if 0 > root_quat.dot( root_pose_bone.rotation_quaternion ): root_quat *= -1
	root_pose_bone.rotation_quaternion = root_quat
	
	pelvis_pose_bone=GetBone( pose_bones, 1 )
	pelvis_quat = Quaternion( ( 0.5, -0.5, -0.5, -0.5 ) )
	pelvis_quat = pelvis_pose_bone.bone.matrix.to_quaternion().inverted() * pelvis_quat
	if 0 > pelvis_quat.dot( pelvis_pose_bone.rotation_quaternion ): pelvis_quat *= -1
	pelvis_pose_bone.rotation_quaternion = pelvis_quat

def split_quat( base_quat, str_plane ):  # get rotation offset for root bone. if str_plane is "NONE", offset is Quaternion( (1,0,0,0) )
	plane_quat = Quaternion( (1,0,0,0) )
	axis_quat  = Quaternion( (1,0,0,0) )
	if "NONE" != str_plane:
		plane_quat = base_quat
		if "ALL" != str_plane:
			if   "XY" == str_plane: axis_vec = Vector( ( 0, 0, 1 ) )
			elif "YZ" == str_plane: axis_vec = Vector( ( 1, 0, 0 ) )
			elif "ZX" == str_plane: axis_vec = Vector( ( 0, 1, 0 ) )
			
			axis_vec_rotated = axis_vec.copy()
			axis_vec_rotated.rotate( base_quat )
			cross = axis_vec.cross( axis_vec_rotated )
			dot = axis_vec.dot( axis_vec_rotated )
			if dot < 1 :
				w = sqrt( ( 1 + dot ) / 2 )
				v = cross.normalized() * sqrt( ( 1 - dot ) / 2 )
				axis_quat = Quaternion( ( w, v.x, v.y, v.z ) )
				plane_quat =  axis_quat.inverted() * base_quat
				
				v.rotate( plane_quat.inverted() )
				axis_quat = Quaternion( ( w, v.x, v.y, v.z ) )
				plane_quat = base_quat * axis_quat.inverted()
				
	return axis_quat, plane_quat
	
def split_quat_test():
	str_plane = bpy.context.scene.gta_tools.ifp_props.root_to_arm_rot
	aobj=bpy.context.active_object
	pose=aobj.pose
	pose_bones=pose.bones
	root_pose_bone=GetBone( pose_bones, 0 )
	base_quat = root_pose_bone.rotation_quaternion
	( axis_quat, plane_quat ) = split_quat( base_quat, str_plane )
	print( "----------" )
	print( "Root ot Armature/ROT : ", str_plane )
	print( "---base_quat---" )
	print( base_quat )
	print( "---axis_quat * plane_quat---" )
	print( axis_quat * plane_quat )
	print( "---axis_quat---" )
	print( axis_quat )
	print( "---plane_quat---" )
	print( plane_quat )
	
def read_tstring( file ):
	tstr = ""
	while True:
		buff = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]
		tstr += buff
		if 4 > len( buff ): break
	return tstr

def get_bonetag( obj_name ):
	inistr = "(tag"
	finstr = ")"
	
	s = obj_name.split( inistr )
	
	if 1 < len( s ):
		ss = s[1].split( finstr )
		if 1 < len( ss ):
			try:
				return int(ss[0])
			except:
				pass
	
	return -1

def GetFCurve( action, target, data_path, array_index, new_flg ):
	## new_flg: if fc is not exist,
	##   False : return None
	##   True  : return New FCurve
	attr = target.path_from_id( data_path )
	for fc in action.fcurves:
		if fc.data_path == attr:
			if fc.array_index == array_index:
				return fc
	if new_flg:
		if target.id_data == target:
			if   data_path == "location":
				target_name = "Location"
			elif data_path == "rotation_quaternion":
				target_name = "Rotation"
		else:
			target_name = target.name
		return action.fcurves.new( attr, index = array_index, action_group = target_name )
	else:
		return None

def SetFCurve( action, target, data_path, array_index, seq, anim_ini, anim_fin ):
	fc = GetFCurve( action, target, data_path, array_index, False )
	
	## ReCalc Sequence
	if None != fc:
		seq_before = []
		seq_after  = []
		buff = [0.0]*len( fc.keyframe_points)*2
		fc.keyframe_points.foreach_get( "co", buff )
		for iseq in range( len( buff ) //2 ):
			if anim_ini > buff[iseq*2]:
				seq_before += [ buff[iseq*2], buff[iseq*2+1] ]
			elif anim_fin < buff[iseq*2]:
				seq_after  += [ buff[iseq*2], buff[iseq*2+1] ]
		seq = seq_before + seq + seq_after
		action.fcurves.remove(fc)
	
	## Set Sequence
	if 0 < len( seq ):
		fin_seq = len( seq ) - 2
		fc = GetFCurve( action, target, data_path, array_index, True )
		if 0 < fin_seq:
			fc.keyframe_points.add( fin_seq //2 )
			fc.keyframe_points.foreach_set( "co", seq[:fin_seq] )
		fc.keyframe_points.insert( seq[fin_seq], seq[fin_seq + 1 ] )

def SetConnectKeys():
	ifp_props=bpy.context.scene.gta_tools.ifp_props
	aobj = bpy.context.active_object
	root = GetBone( aobj.pose.bones, 0 )
	frams = []
	aobj_buff = []
	root_buff = []
	
	## set frames
	frams.append( bpy.context.scene.frame_current - 1 )
	frams.append( bpy.context.scene.frame_current )
	
	## set keys to buffer
	for ifram,fram in enumerate( frams ):
		bpy.context.scene.frame_set( fram )
		aobj_buff.append([])
		root_buff.append([])
		aobj_buff[ifram].append( aobj.location.copy() )
		aobj_buff[ifram].append( aobj.rotation_quaternion.copy() )
		root_buff[ifram].append( root.location.copy() )
		root_buff[ifram].append( root.rotation_quaternion.copy() )
	
	## calc swapped locations
	aobj_swapped = aobj_buff[1][0].copy()
	root_swapped = root_buff[1][0].copy()
	root_swapped.rotate( root.bone.matrix.to_quaternion() )
	root_swapped.rotate( aobj_buff[1][1] )
	#aobj_swapped.rotate( aobj_buff[1][1] )
	pos_set  = False
	pos_swap = False
	for ixyz in range( 3 ):
		if ifp_props.set_connect_keys_pos[ixyz]:
			pos_set = True
		if ifp_props.set_connect_keys_pos_swap[ixyz]:
			aobj_swapped[ixyz] = root_swapped[ixyz] + aobj_buff[1][0][ixyz]
			root_swapped[ixyz] = 0
			pos_swap = True
	root_swapped.rotate( root.bone.matrix.to_quaternion().inverted() )
	
	## set keys at previous frame
	bpy.context.scene.frame_set( frams[0] )
	for ixyz in range( 3 ):
		if ifp_props.set_connect_keys_pos[ixyz] or ifp_props.set_connect_keys_pos_swap[ixyz]:
			aobj.keyframe_insert(data_path="location", index = ixyz, group = "Location" )
	if pos_set or pos_swap:
		root.keyframe_insert( data_path="location", group = root.name )
	if ifp_props.set_connect_keys_rot or "NONE" != ifp_props.set_connect_keys_rot_swap:
		aobj.keyframe_insert( data_path="rotation_quaternion", group = "Rotation" )
		root.keyframe_insert( data_path="rotation_quaternion", group = root.name )
	
	## move root-keys to armature, and set keys at current frame
	bpy.context.scene.frame_set( frams[1] )
	for ixyz in range( 3 ):
		if ifp_props.set_connect_keys_pos_swap[ixyz]:
			aobj.location[ixyz] = aobj_swapped[ixyz]
			aobj.keyframe_insert(data_path="location", index = ixyz, group = "Location" )
		elif ifp_props.set_connect_keys_pos[ixyz]:
			aobj.keyframe_insert(data_path="location", index = ixyz, group = "Location" )
	if pos_swap:
		root.location = root_swapped
		root.keyframe_insert(data_path="location", group = root.name )
	elif pos_set:
		root.keyframe_insert(data_path="location", group = root.name )
	
	if "NONE" != ifp_props.set_connect_keys_rot_swap:
		default_root_quat=Quaternion( ( sqrt(1/2), 0, 0, sqrt(1/2) ) )
		root_buff[1][1] = root.bone.matrix.to_quaternion() * root_buff[1][1]
		( axis_quat, plane_quat ) = split_quat( root_buff[1][1], ifp_props.set_connect_keys_rot_swap )
		plane_quat = plane_quat * default_root_quat.inverted()
		plane_quat = plane_quat * aobj_buff[1][1]
		aobj.rotation_quaternion = plane_quat
		
		axis_quat = default_root_quat * axis_quat
		axis_quat = root.bone.matrix.to_quaternion().inverted() * axis_quat
		root.rotation_quaternion = axis_quat
		
		aobj.keyframe_insert(data_path="rotation_quaternion", group = "Rotation" )
		root.keyframe_insert(data_path="rotation_quaternion", group = root.name )
		
	elif ifp_props.set_connect_keys_rot:
		aobj.keyframe_insert( data_path="rotation_quaternion", group = "Rotation" )
		root.keyframe_insert( data_path="rotation_quaternion", group = root.name )
	
	## clear following anim data after Current Frame
	## under construction
	#if ifp_props.clear_following:
	#	
	
	## set linear interpolation prop to existing previous frame
	if ifp_props.linear_interpolation:
		for ixyz in range( 3 ):
			if ifp_props.set_connect_keys_pos[ixyz] or ifp_props.set_connect_keys_pos_swap[ixyz]:
				set_linear_interpolation( aobj.animation_data.action, aobj , "location", ixyz, frams )
			if pos_set or pos_swap:
				set_linear_interpolation( aobj.animation_data.action, root , "location", ixyz, frams )
		for iwxyz in range( 4 ):
			if ifp_props.set_connect_keys_rot or "NONE" != ifp_props.set_connect_keys_rot_swap:
				set_linear_interpolation( aobj.animation_data.action, aobj , "rotation_quaternion", iwxyz, frams )
				set_linear_interpolation( aobj.animation_data.action, root , "rotation_quaternion", iwxyz, frams )

def set_linear_interpolation( action, target, data_path, array_index, frams ):
	fc = GetFCurve( action, target, data_path, array_index, False )
	if None != fc:
		kfs = [ None, None, None ]
		for kf in fc.keyframe_points:
			if kf.co[0] < frams[0]:
				kfs[0] = kf
			else:
				if kf.co[0] == frams[0]:
					kfs[1] = kf
				elif kf.co[0] == frams[1]:
					kfs[2] = kf
					break
		for kf in kfs:
			if None != kf:
				#print( target, data_path, array_index, kf.co[0] )  ##for Debug
				kf.interpolation = "LINEAR"

def clear_anim( action, target, data_path, array_index, anim_ini, anim_fin ):
	seq = []
	fc = GetFCurve( action, target, data_path, array_index, False )
	if None != fc:
		buff = [0.0]*len( fc.keyframe_points)*2
		fc.keyframe_points.foreach_get( "co", buff )
		for iseq in range( len( buff ) //2 ):
			if anim_ini > buff[iseq*2] or anim_fin < buff[iseq*2]:
				seq += [ buff[iseq*2], buff[iseq*2+1] ]
		action.fcurves.remove(fc)
	if 0 < len( seq ):
		fin_seq = len( seq ) - 2
		fc = GetFCurve( action, target, data_path, array_index, True )
		fc.keyframe_points.add( fin_seq // 2 )
		fc.keyframe_points.foreach_set( "co", seq[:fin_seq] )
		fc.keyframe_points.insert( seq[fin_seq], seq[fin_seq + 1 ] )



def ImportIFP( flags ):  # flags(0x1):  False: read list, True: read active anim
	time_ini=time.clock()
	
	ifp_props=bpy.context.scene.gta_tools.ifp_props
	read_active = ( 0x1 and flags )
	
	global script_info
	print( "\n-----\n" + script_info + "\n-----" )
	if read_active:
		print( "Mode: Load Selected Anim( ID: "+str( ifp_props.active_anim_id )+" )" )
	else:
		print( "Mode: Read List" )
	
	## Open File
	try:
		file=open( ifp_props.filepath, "rb" )
	except:
		print( "-----\nOpen File Error: "+ifp_props.filepath )
		return
	
	print( "Source IFP File : "+ifp_props.filepath )
	
	## Read Anim
	# Header
	# identifier  : CHAR[4]
	# offset      : UINT              // offset to the end of the section
	# // Note: Offsets are always relative to the current file position.
	
	Anims=[]
	fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]
	end_of_file = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
	#print( fourcc, end_of_file )
	
	if "ANPK" == fourcc:
		# "ANPK"
		# information : INFO<TAnimation>  // section information data (see below)
		
		# "INFO"
		# entries     : INT32             // (count of encapsuled sections)
		# name        : TString           // (name of the collection)
		# sections    : T[entries]        // Subsections
		
		fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]  # "INFO"
		end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
		#print( fourcc, hex(end_of_section) )
		
		num_anims = struct.unpack( "<i", file.read( 4 ) )[0]
		internal_file_name = read_tstring( file )
		
		if False == read_active:
			ifp_props.active_anim_id = -1
			ifp_props.anims_clear()
			ifp_props.ifp_name = internal_file_name
		
		print( "IFP Name : "+ifp_props.ifp_name )
		print( "Number of Anims : ", num_anims )
		
		for ianim in range( num_anims ):
			# "NAME"
			# name        : TString
			fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]   # "NAME"
			end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
			anim_name = read_tstring( file )
			print( " ", anim_name )
			
			# "DGAN"
			# anim. info  : INFO<CPAN>
			fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]   # "DGAN"
			end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
			end_of_dgan = end_of_section
			
			# "INFO"
			# entries     : INT32             // (count of encapsuled sections)
			# name        : TString           // (name of the collection)
			# sections    : T[entries]        // Subsections
			
			fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]   # "INFO"
			end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
			#print( fourcc, hex(end_of_section) )
			
			num_anims = struct.unpack( "<i", file.read( 4 ) )[0]
			file.seek( end_of_section, 0 )
			
			if read_active:
				if ifp_props.active_anim_id == ianim:
					anim = ClassIFPAnim()
					anim.name = anim_name
					
					for ianim in range( num_anims ):
						# "CPAN"
						# object info : ANIM
						fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]   # "CPAN"
						end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
						#print( fourcc, hex(end_of_section) )
						
						# "ANIM"
						# object name : TString           // Also the name of the bone.
						# // Note: Because of this fact that this string uses 28 bytes by default.
						# frames      : INT32             // Number of frames
						# unknown     : INT32             // Usually 0
						# next        : INT32             // Next sibling
						# prev        : INT32             // Previous sibling
						# frame data  : KFRM
						# 
						### ???
						# prev    --> bone ID
						
						fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]   # "ANIM"
						end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
						#print( fourcc, hex(end_of_section) )
						
						obj_name = bytes.decode( file.read( 28 ), errors='replace' ).split( '\0' )[0]
						data = struct.unpack( "<4i", file.read( 16 ) )
						num_frams = data[0]
						bone_id = data[3]
						
						# frame data  : KFRM              // There are 3 known specialisations of this base structure:
						#                                 //   KR00, KRT0 and KRTS.
						fourcc = bytes.decode( file.read( 4 ), errors='replace' ).split( '\0' )[0]
						end_of_section = struct.unpack( "<i", file.read( 4 ) )[0] + file.tell()
						end_of_kfrm = end_of_section
						#print( fourcc, hex(end_of_section) )
						
						#if "KR00" != fourcc and "KRT0" != fourcc and "KRTS" != fourcc: break
						
						obj = ClassIFPObj()
						obj.name = obj_name
						obj.bid = bone_id
						obj.kfrm = fourcc
						
						for ifram in range( num_frams ):
							fram = ClassIFPFram()
							if   fourcc == "KR00":
								# "KR00"
								# rot         : TVector4[X]       // quaternion rotation (float x,y,z,w)
								data = struct.unpack( "<4f", file.read( 16 ) )
								fram.rot = Quaternion( ( float( data[3] ), float( data[0] ), float( data[1] ), float( data[2] ) ) ).inverted()
							
							elif fourcc == "KRT0":
								# "KRT0"
								# {
								#   rot       : TVector4          // quaternion rotation (float x,y,z,w)
								#   pos       : TVector3          // Translation
								# }[X]                            // Repeated X times.
								data = struct.unpack( "<7f", file.read( 28 ) )
								fram.rot = Quaternion( ( float( data[3] ), float( data[0] ), float( data[1] ), float( data[2] ) ) ).inverted()
								fram.pos = Vector    ( ( float( data[4] ), float( data[5] ), float( data[6] ) ) )
							
							elif fourcc == "KRTS":
								# "KRTS"
								# {
								#   rot       : TVector4          // quaternion rotation (float x,y,z,w)
								#   pos       : TVector3          // Translation
								#   scale     : TVector3          // Scale
								# }[X]                            // Repeated X times.					
								data = struct.unpack( "<10f", file.read( 40 ) )
								fram.rot   = Quaternion( ( float( data[3] ), float( data[0] ), float( data[1] ), float( data[2] ) ) ).inverted()
								fram.pos   = Vector    ( ( float( data[4] ), float( data[5] ), float( data[6] ) ) )
								fram.scale = Vector    ( ( float( data[7] ), float( data[8] ), float( data[9] ) ) )
							
							# time key    : FLOAT             // This value is the last one of the specialised sections.
							# --> convert to frame index, temporally 25 FPS
							data = struct.unpack( "<f", file.read( 4 ) )
							#if 100 < int( round( data[0] * 25 ) ):
							#	file.seek( end_of_kfrm, 0 )
							#	break
							fram.time = int( round( data[0] * 25 ) )
							#print( fram.time, data[0] )
							obj.frams.append( fram )
						anim.objs.append( obj )
					Anims.append(anim)
			
			else:
				ifp_props.anims.add().name=anim_name
			
			file.seek( end_of_dgan, 0 )
	
	elif "ANP3" == fourcc:
		# ANP3 Header
		# 4b   - FourCC   - 'ANP3' (Animation Package 3, Version identifier. However there is no pack with ANP2)
		# 4b   - Int32    - Offset to end of file
		# 24b  - Char[24] - internal file name used in the script
		# 4b   - Int32    - Number of Animations		
		
		internal_file_name = bytes.decode( file.read( 24 ), errors='replace' ).split( '\0' )[0]
		data = struct.unpack( "<i", file.read( 4 ) )
		num_anims=data[0]
		#print( endof_file, internal_file_name, num_anims )
		
		if False == read_active:
			ifp_props.active_anim_id=-1
			ifp_props.anims_clear()
			ifp_props.ifp_name = internal_file_name
		
		print( "IFP Name : "+ifp_props.ifp_name )
		print( "Number of Anims : ", num_anims )
		
		for ianim in range( num_anims ):
			# 24b  - Char[24] - Animation Name
			# 4b   - Int32    - Number of Objects
			# 4b   - Int32    - Size of frame data
			# 4b   - Int32    - Unknown, always 1
			
			anim_name=bytes.decode( file.read( 24 ), errors='replace' ).split( '\0' )[0]
			data=struct.unpack( "<3i", file.read( 12 ) )
			num_objs=data[0]
			size_fram=data[1]
			unk=data[2]
			
			if read_active:
				anim = ClassIFPAnim()
				anim.name = anim_name
			else:
				#print( ianim, anim_name, num_objs, size_fram, unk )  ## for Debug
				print( " ", anim_name )
				ifp_props.anims.add().name=anim_name
			
			for iobj in range( num_objs ):
				# 24b  - Char[24] - Object Name
				# 4b   - Int32    - Frame type: Child=3, Root=4
				# 4b   - Int32    - Number of Frames
				# 4b   - Int32    - Bone ID
				
				obj_name=bytes.decode( file.read( 24 ), errors='replace'  ).split( '\0' )[0]
				data=struct.unpack( "<3i", file.read( 12 ) )
				type_fram=data[0]
				num_frams=data[1]
				bone_id=data[2]
				#print( obj_name, bone_id, type_fram, num_frams, bone_id )
				
				if read_active and ( ifp_props.active_anim_id == ianim ):
					obj = ClassIFPObj()
					obj.name = obj_name
					obj.bid = bone_id
					if   3 == type_fram: obj.kfrm = "KR00"
					elif 4 == type_fram: obj.kfrm = "KRT0"
					
					for ifram in range( num_frams ):
						# 2b   - Int16    - Quaternion X
						# 2b   - Int16    - Quaternion Y
						# 2b   - Int16    - Quaternion Z
						# 2b   - Int16    - Quaternion W
						# 2b   - Int16    - Time (in seconds)
						# 2b   - Int16    - Translation X
						# 2b   - Int16    - Translation Y
						# 2b   - Int16    - Translation Z
						
						fram = ClassIFPFram()
						data=struct.unpack( "<5h", file.read( 10 ) )
						fram.rot = Quaternion( ( float( data[3] )/4096, float( data[0] )/4096, float( data[1] )/4096, float( data[2] )/4096 ) )
						fram.time = data[4] // 2
						
						if type_fram == 4:
							data=struct.unpack( "<3h", file.read( 6 ) )
							fram.pos = Vector( ( float( data[0] ), float( data[1] ), float( data[2] ) ) ) /1024
						
						obj.frams.append( fram )
					anim.objs.append( obj )
				else:
					if 4 == type_fram: file.seek( num_frams * 16 , 1 )
					else             : file.seek( num_frams * 10 , 1 )
			if read_active and ( ifp_props.active_anim_id == ianim ):
				Anims.append(anim)
	else:
		print( "-----\nHeader Error"+ifp_props.filepath )
		file.close()
		return
	
	file.close()
	
	
	#####
	# Assing Anim Data to Active Armature
	
	if read_active:
		aobj=bpy.context.active_object
		if aobj.mode != 'POSE':
			print( "-----\nMode Error: "+aobj.name+" (mode: "+aobj.mode+")" )
			return
		print( "-----\nTarget Object: "+aobj.name+" (mode: "+aobj.mode+")" )
		arm=aobj.data
		if aobj.type != 'ARMATURE':
			print( "-----\nObject Type Error: "+aobj.data.name+" (type: "+aobj.type+")" )
			return
		print( "Object Type: "+aobj.data.name+" (type: "+aobj.type+")" )
		pose=aobj.pose
		pose_bones=pose.bones
		print( "Number of Pose Bones: "+str( len( pose_bones ) ) )
		
		## option flags
		skip_root_flg = False
		if ifp_props.skip_pos and (True, True, True) == ifp_props.skip_root[:]:
			skip_root_flg = True
		
		root_to_arm_pos_flg = False
		if ifp_props.root_to_arm and not (False, False, False) == ifp_props.root_to_arm_pos[:]:
			root_to_arm_pos_flg = True
		
		root_to_arm_pos_root_flg = False
		if not ifp_props.root_to_arm and not (True, True, True) == ifp_props.root_to_arm_pos[:]:
			root_to_arm_pos_root_flg = True
		print( root_to_arm_pos_root_flg )
		
		if ifp_props.reset_anim:
			ResetAnim()
			ResetPose()
			if ifp_props.root_to_arm:
				if ifp_props.reset_arm:
					ResetArmature()
			#AnimDirection()
		
		## set up for Frames
		anim=Anims[0]
		ini_fram=bpy.context.scene.frame_current
		min_fram=0
		max_fram=0
		for obj in anim.objs:
			for fram in obj.frams:
				min_fram=min( [min_fram, fram.time] )
				max_fram=max( [max_fram, fram.time] )
		start_fram=ini_fram+min_fram
		end_fram  =ini_fram+max_fram
		
		bpy.ops.object.mode_set( mode='POSE' )
		if ifp_props.adjust_frame_range:
			if None != aobj.animation_data:
				print( aobj.animation_data )
				bpy.context.scene.frame_start = min( [ start_fram, aobj.animation_data.action.frame_range[0] ] )
				bpy.context.scene.frame_end   = max( [ end_fram,   aobj.animation_data.action.frame_range[1] ] )
			else:
				bpy.context.scene.frame_start = start_fram
				bpy.context.scene.frame_end = end_fram
		bpy.context.scene.frame_current = ini_fram
		
		print( "-----")
		print( "Anim Name    : " + anim.name )
		print( "Anim Start   : " + str(start_fram) )
		print( "Anim End     : " + str(end_fram) )
		print( "-----" )
		
		if None == aobj.animation_data:
			aobj.animation_data_create()
			action = bpy.data.actions.new( aobj.name )
			aobj.animation_data.action = action
		
		#####
		## Assing Anim Data
		root_pose_bone=GetBone( pose_bones, 0 )
		default_root_quat = Quaternion( ( sqrt(1/2), 0, 0, sqrt(1/2) ) )
		aobj_quat_ini = aobj.rotation_quaternion.copy()
		aobj_co_ini = aobj.location.copy()
		root_quat_ofs = Quaternion( ( 1, 0, 0, 0 ) )
		root_co_ofs = Vector( (0,0,0) )
		
		## get root co/quat at initial frame of loaded ifp 
		ifp_quat_ini = default_root_quat
		ifp_co_ini   = Vector( ( 0, 0, 0 ) )
		for obj in anim.objs:
			## for Debug
			if 0 == obj.bid:
				for fram in obj.frams:
					if 0 ==  fram.time:
						ifp_quat_ini = fram.rot
						ifp_co_ini   = fram.pos
						break
				break
		
		## calc offset POS/ROT of Root Bone, and split
		if ifp_props.use_current_root:
			if 0 < len( aobj.animation_data.action.fcurves ):
				## ROT
				if "NONE" != ifp_props.use_current_root_rot:
					root_quat_ofs = root_pose_bone.matrix.to_quaternion().copy() * ifp_quat_ini.inverted()
					( axis_quat, root_quat_ofs ) = split_quat ( root_quat_ofs, ifp_props.use_current_root_rot )
				## POS
				offset_vec = root_pose_bone.location.copy()
				offset_vec.rotate( root_pose_bone.bone.matrix.to_quaternion() )
				offset_vec = offset_vec - ifp_co_ini
				for ixyz in range( 3 ):
					if ifp_props.use_current_root_pos[ixyz]:
						root_co_ofs[ixyz] += offset_vec[ixyz]
		
		## set Anim Data
		for obj in anim.objs:
			## for Debug
			print( obj.name,":", obj.bid, obj.kfrm, len( obj.frams ) )
			
			## get Root-Bone entry
			pose_bone=GetBone( pose_bones, obj.bid )
			if None == pose_bone:
				pose_bone=GetBoneByName( pose_bones, obj.name )
			if None == pose_bone:
				print( "  not found:", obj.name, obj.bid )
				continue
			
			## get Anim-Range for FCurves
			anim_ini = float( obj.frams[0].time  + ini_fram )
			anim_fin = float( obj.frams[-1].time + ini_fram )
			
			## set Position
			if "KR00" != obj.kfrm:
				## Params foe setting POS
				aseq = [ [], [], [] ]
				bseq = [ [], [], [] ]
				
				### set Init Root Keys if "init_root_key_pos" enebled
				#if root_pose_bone == pose_bone:
				#	print( obj.name,":", obj.bid, obj.kfrm, len( obj.frams ) )
				#	print( obj.frams[0].time )
				#	for ixyz in range( 3 ):
				#		if ifp_props.init_root_key_pos[ixyz]:
				#			print( obj.frams[0].pos[ixyz] )
				
				for fram in obj.frams:
					fram_time = float( fram.time + ini_fram )
					pos = Vector ( fram.pos )
					
					## split Root-POS into Armature and Root-Bone
					if root_pose_bone == pose_bone:
						apos = Vector( ( 0, 0, 0 ) )
						bpos = Vector( ( 0, 0, 0 ) )
						for ixyz in range( 3 ):
							if not ifp_props.skip_pos or not ifp_props.skip_root[ixyz]:
								if ifp_props.root_to_arm and ifp_props.root_to_arm_pos[ixyz]:
									apos[ixyz] += pos[ixyz] - ifp_co_ini[ixyz]
									bpos[ixyz] += ifp_co_ini[ixyz]
								else:
									bpos[ixyz] += pos[ixyz]
						apos.rotate( aobj_quat_ini )
						apos += aobj_co_ini
						bpos.rotate( root_quat_ofs )
						bpos += root_co_ofs
						bpos.rotate( pose_bone.bone.matrix.to_quaternion().inverted() )
						
						## set POS sequences for FCurves
						for ixyz in range( 3 ):
							aseq[ixyz] += [ fram_time, apos[ixyz] ]
							bseq[ixyz] += [ fram_time, bpos[ixyz] ]
					
					## set POS sequences for FCurves of child bones
					elif not ifp_props.skip_pos or not ifp_props.skip_children:
						pos.rotate( pose_bone.bone.matrix.to_quaternion().inverted() )
						for ixyz in range( 3 ):
							bseq[ixyz] += [ fram_time, pos[ixyz] ]
				
				### remove All-Zero Sequence
				#for ixyz in range( 3 ):
				#	all_zero = True
				#	for iseq in range( len( aseq[ixyz] ) //2 ):
				#		data = aseq[ixyz][iseq*2+1]
				#		if 0.001 < abs( data ):
				#			all_zero = False
				#			break
				#	if all_zero:
				#		aseq[ixyz][:] = []
				#	
				#	all_zero = True
				#	for iseq in range( len( bseq[ixyz] ) //2 ):
				#		data = bseq[ixyz][iseq*2+1]
				#		print( data )
				#		if 0.001 < abs( data ):
				#			all_zero = False
				#			break
				#	if all_zero:
				#		bseq[ixyz][:] = []
					
				## set POS sequences to FCurves
				if root_pose_bone == pose_bone:
					if not skip_root_flg:
						if root_to_arm_pos_flg:
							for ixyz in range( 3 ):
								SetFCurve( aobj.animation_data.action, aobj     , "location", ixyz, aseq[ixyz], anim_ini, anim_fin )
						if root_to_arm_pos_root_flg:
							for ixyz in range( 3 ):
								SetFCurve( aobj.animation_data.action, pose_bone, "location", ixyz, bseq[ixyz], anim_ini, anim_fin )
				else:
					for ixyz in range( 3 ):
						SetFCurve( aobj.animation_data.action, pose_bone, "location", ixyz, bseq[ixyz], anim_ini, anim_fin )
			
			## params for Rotation settings
			aseq = [ [], [], [], [] ]
			bseq = [ [], [], [], [] ]
			prev_aquat = aobj.rotation_quaternion
			prev_bquat = pose_bone.rotation_quaternion
			
			## set Rotation
			for fram in obj.frams:
				fram_time = float( fram.time + ini_fram )
				aquat = Quaternion( ( 1, 0, 0, 0 ) )
				bquat = Quaternion( fram.rot )
				
				if root_pose_bone == pose_bone:
					## Split Root Rotation into Armature and RootBone
					if ifp_props.root_to_arm and "NONE" != ifp_props.root_to_arm_rot:
						bquat = bquat * ifp_quat_ini.inverted()
						( bquat, aquat ) = split_quat( bquat, ifp_props.root_to_arm_rot )
						aquat = aobj_quat_ini * aquat
						bquat = bquat * ifp_quat_ini
					if ifp_props.use_current_root and "NONE" != ifp_props.use_current_root_rot:
						bquat = root_quat_ofs * bquat
				bquat = pose_bone.bone.matrix.to_quaternion().inverted() * bquat
				
				## Fix 360deg-skipped Rotation
				if 0 > aquat.dot( prev_aquat ): aquat *= -1
				prev_aquat = aquat
				
				if 0 > bquat.dot( prev_bquat ): bquat *= -1
				prev_bquat = bquat
				
				## append rotation sequences for FCurves
				for iwxyz in range( 4 ):
					if root_pose_bone == pose_bone:
						if ifp_props.root_to_arm and "NONE" != ifp_props.root_to_arm_rot:
							aseq[iwxyz] += [ fram_time, aquat[iwxyz] ]
					bseq[iwxyz] += [ fram_time, bquat[iwxyz] ]
			
			## set rotation sequences to FCurves
			for iwxyz in range( 4 ):
				if root_pose_bone == pose_bone and ifp_props.root_to_arm and "NONE" != ifp_props.root_to_arm_rot:
					SetFCurve( aobj.animation_data.action, aobj     , "rotation_quaternion", iwxyz, aseq[iwxyz], anim_ini, anim_fin )
					SetFCurve( aobj.animation_data.action, pose_bone, "rotation_quaternion", iwxyz, bseq[iwxyz], anim_ini, anim_fin )
				else:
					SetFCurve( aobj.animation_data.action, pose_bone, "rotation_quaternion", iwxyz, bseq[iwxyz], anim_ini, anim_fin )
		
		bpy.context.scene.frame_current=end_fram
		#if ifp_props.auto_connect_key: SetConnectKeys()
		ifp_props.exp_internal_file_name = internal_file_name
		ifp_props.exp_anim_name = anim.name
	
	print( "-----\nFinished : Successfully." )
	print( "Elapsed Time : "+str( round( time.clock() - time_ini, 3 ) )+" sec\n-----\n" )
	return



